#!/bin/bash

shellfu import dist
shellfu import jat_dump

#
# Keep and compare system states
#

#
# State dir root
#
STATE__DIR=${STATE__DIR:-/var/tmp/state}

state__scan() {
    #
    # Scan and save state using modules $@
    #
    # Usage:
    #    state__scan [MODULE]..
    #
    # Initiate and perform scan using specified MODULEs, or if no
    # MODULE is specified, using all available modules.  Scan
    # results are saved under $STATE__DIR and
    # you can use utility function state__diff2()
    # to compare last two scans.
    #
    # Example:
    #
    #     state__scan
    #     # do something that should not change state
    #     state__scan
    #     state__diff2     # FAIL if state had changed
    #
    local m_todo=()     # modules to run
    local scandir       # directory of this scan
    local scanid        # this scan ID
    local Module        # each module in question
    scanid="scan-$(date +%s-%N)"
    if test -n "$1"; then
        m_todo=("$@")
    else
        mapfile -t m_todo <<<"$(__state__lsmodules)"
    fi
    jat__pstarts -h "initiate state scan: $scanid"
        scandir="$STATE__DIR/scan/$scanid"
        mkdir -p "$scandir/meta" "$scandir/data"
        for Module in "${m_todo[@]}"; do
            echo "$Module"
        done | sort > "$scandir/meta/m_todo"
        hostname > "$scandir/meta/hostname"
    jat__pend
    for Module in "${m_todo[@]}"; do
        __state__pushd "$scandir/data" || return 3
            __state__save1
        __state__popd || return 3
    done
    jat__pstarts -h "submit tarball: $scanid"
        __state__starball scan "$scanid"
    jat__pend
}

state__diff() {
    #
    # Compare two scans' data
    #
    # Usage:
    #     state__diff [-n] [-p PHASE] [-H HANDLER] [-LA LABEL_A] [-LB LABEL_B] SCAN_A SCAN_B
    #
    # Create test phase and assert that scans SCAN_A and SCAN_B are the same.
    #
    # Phase name will be generated, unless PHASE is passed.  LABEL_A and LABEL_B
    # can be passed, which are then used to label the scans in log and phase name,
    # and in handler.
    #
    # If you pass HANDLER, it is treated as function and will be executed once
    # for each SCAN_A and SCAN_B, with workdir set to a copy of the scan data.
    # This gives give you oppportunity to weed out irrelevant data before the
    # diff is done.
    #
    # Handlers can distinguish between the particular runs by calling
    # state__hget() function.
    #
    # Pass -n to ensure the diff assert always passes; the resulting phase will be
    # purely informational.
    #
    local scan_a        # SCAN_A
    local scan_b        # SCAN_B
    local label_a       # custom label for SCAN_A
    local label_b       # custom label for SCAN_B
    local diff_dir       # dedicated directory for diff
    local pname         # phase name
    local Handler=true  # modification handler
    local o_es=0        # diff expected exit status
    while true; do case $1 in
        -LA) label_a=$2; shift 2 ;;
        -LB) label_b=$2; shift 2 ;;
        -n)  o_es=0,1; shift ;;
        -p) pname=$2; shift 2 ;;
        -H) Handler=$2; shift 2 ;;
        -*) __state__usage diff; return 3 ;;
        *) break ;;
    esac done
    scan_a=$1; scan_b=$2; shift 2
    test -d "$STATE__DIR/scan/$scan_a" || { jat__log_error "no such scan: $scan_a"; return 3; }
    test -d "$STATE__DIR/scan/$scan_b" || { jat__log_error "no such scan: $scan_b"; return 3; }
    test -n "$label_a" || label_a=$scan_a
    test -n "$label_b" || label_b=$scan_b
    test -n "$pname" || pname="compare scans: $label_a, $label_b"
    jat__pstartt -h "$pname"
        diff_id="diff-$(date +%s-%N)"
        test "$label_a" != "$scan_a" -o "$label_b" != "$scan_b" && {
            jat__log_info \
                "using label '$label_a' for scan id $scan_a" \
                "using label '$label_b' for scan id $scan_b"
        }
        diff_dir="$STATE__DIR/diff/$diff_id"
        mkdir -p "$diff_dir"
        __state__pushd "$diff_dir" || return 3
            __state__handle SCAN_A "$scan_a" "$label_a"
            __state__handle SCAN_B "$scan_b" "$label_b"
            jat__cmd -h "Run full diff on original data" \
                -S 0,1 \
                -o original.diff \
                __state__diff \
                    "$scan_a-original" "$scan_b-original" \
                    "$label_a" "$label_b"
            jat__cmd -h "Run full diff on mangled data" \
                -S "$o_es" \
                -o mangled.diff \
                __state__diff \
                    "$scan_a-mangled" "$scan_b-mangled" \
                    "$label_a" "$label_b"
            jat_dump__file "mangled.diff"
            jat__submit "mangled.diff" "$diff_id-mangled.diff"
        __state__popd || return 3
        __state__starball diff "$diff_id"
    jat__pend
}

state__diff_last2() {
    #
    # Compare last 2 scans' data
    #
    # Usage:
    #     state__diff_last2 [diff-options]
    #
    # Same as state__diff(), but SCAN_A and SCAN_B arguments are
    # automatically taken to be "last 2 scans".  Other options have the same
    # meaning.
    #
    local pname         # phase name
    local opts=("$@")   # options
    state__diff \
      -p "compare last two scans" \
      "${opts[@]}" \
      "$(__state__last 2)" \
      "$(__state__last 1)"
}

state__hget() {
    #
    # Get property $1 of currently mangled scan
    #
    # Usage:
    #    state__hget position|label
    #
    # Call this from handler to detect whether the handler is being
    # called on SCAN_A or SCAN_B.  Calling with 'position' works
    # always and prints literally 'SCAN_A' or 'SCAN_B'.  'label'
    # only works if labels have been set using -LA and -LB options;
    # in that case respective labels are printed.
    #
    local prop=$1
    test -n "$__state__handled_position" || {
        jat__log_error \
            "!! bad usage of state__hget() !!" \
            "this function can be only called from within a diff" \
            "handler specified via -H option"
        return 2
    }
    case $prop in
        position)   echo "$__state__handled_position" ;;
        label)      echo "$__state__handled_label" ;;
    esac
}

#          #                            State of your system is in peril #
# INTERNAL # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
#          #                 ...if you use anything from below this line #

__state__diff() {
    #
    # Run diff on dirs $1 and $2 decorated with labels $3 and $4
    #
    # Usage:
    #     __state__diff DIR_A DIR_B [LABEL_A LABEL_B]
    #
    # Run diff, ignoring internal non-deterministic data and
    # decorating with labels, if provided.
    #
    local dir_a=$1
    local dir_b=$2
    local label_a=$3
    local label_b=$4
    local pfx_a
    local pfx_b
    local tmp
    local es=0
    pfx_a=$(__state__repfx "$label_a" "$label_b" | sed -n 1p)
    pfx_b=$(__state__repfx "$label_a" "$label_b" | sed -n 2p)
    tmp=$(mktemp -t __state__diff.XXXXXXXX)
    diff -u -x diag -r "$dir_a" "$dir_b" >"$tmp"; es=$?
    while read -r line; do case $line in
        ---" $dir_a"*) echo "--- $pfx_a${line#--- }" ;;
        +++" $dir_b"*) echo "+++ $pfx_b${line#+++ }" ;;
        *)   echo "$line" ;;
    esac done <"$tmp"
    rm "$tmp"
    return $es
}

__state__repfx() {
    #
    # Right-pad label prefixes
    #
    local a=$1
    local b=$2
    local w=0
    test -z "$a" && test -z "$b" && return 0
    test -n "$a" && pa="[$a] "
    test -n "$b" && pb="[$b] "
    test ${#pa} -gt "$w" && w=${#pa}
    test ${#pb} -gt "$w" && w=${#pb}
    printf "%-${w}s\n" "$pa" "$pb"
}

__state__handle() {
    #
    # Handle scan $1 as $2 using handler function $Handler
    #
    local position=$1
    local id=$2
    local label=$3
    local scanpath="$STATE__DIR/scan/$id"
    local hint="scan '$id'"
    test "$id" == "$label" || hint="scan '$label' ('$id')"
    test -n "$Handler" || return 0
    jat__log_info "running handler: $Handler on $hint" \
                  "   position='$position'" \
                  "   id='$id'" \
                  "   label='$label'"
    (
        cp -ar "$scanpath" "$id-original" || exit 3
        cp -ar "$scanpath" "$id-mangled"  || exit 3
        cd "$id-mangled" || exit 3
        __state__handled_position=$position
        __state__handled_label=$label
        "$Handler"
    ) || {
        jat__log_error "handler failed: '$Handler' on $hint"
        return 3
    }
}

__state__last() {
    #
    # Get id of last $1-th scan
    #
    local n=$1      # wanted ordinal
    local all=()    # all scans (latest to oldest)
    local yet       # number of ^^
    mapfile -t all <<<"$(__state__lsscans | sort -r)"
    yet=${#all[@]}
    test "$n" -gt 0 || {
        jat__log_error "N must be an integer greater than 0"
        return 2
    }
    test "$yet" -ge "$n" || {
        jat__log_error "there's not enough scans yet: $yet < $n"
        return 2
    }
    echo "${all[$((n-1))]}"
}

__state__lsmodules() {
    #
    # List all available modules
    #
    set \
      | grep '^__state__m__[[:alnum:]_]* () $' \
      | cut -d' ' -f1 \
      | cut -d_ -f7- \
      | sort
}

__state__lsscans() {
    #
    # List all previous scans
    #
    find "$STATE__DIR/scan" \
        -mindepth 1 -maxdepth 1 \
        -name "scan-*" -type d \
        -printf '%P\n' \
      | sort \
      | grep .
}

__state__m__rpmv() {
    #
    # Module: rpm -V
    #
    local action=$1     # what to do
    local pkgs=()       # packages found
    local pkg           # each package name
    case $action in
        diag)
            rpm -qa --dump \
              | sort > rpm_dump.log
            ;;
        scan)
            mapfile -t pkgs <<<"$(rpm --qf '%{NAME}\n' -qa | sort)"
            for pkg in "${pkgs[@]}"; do
                rpm -V "$pkg" \
                  | sed "s/^/$pkg: /"
            done \
              | sort -t/ -k2
            ;;
        describe)
            echo "rpm verification"
            ;;
        version)
            echo "0.0"
            ;;
    esac
}

__state__mu__services_intsys() {
    #
    # Print which init system is used
    #
    dist__test \
        id.vlike == rhel \
        ver.x ge 7 \
     && echo systemd \
     && return 0
    echo sysvinit
}

__state__m__services() {
    #
    # Module: State of services
    #
    local action=$1     # what to do
    local services=()   # services found
    local sname         # each service name
    local sstate        # each service state
    local initsys
    initsys=$(__state__mu__services_intsys)
    #FIXME: make it work for RHEL7+ and across
    case $action in
        diag)
            case $initsys in
                systemd)
                    journalctl --no-pager > journalctl.log
                    ;;
                sysvinit)
                    chkconfig | sort > chkconfig.log
                    ;;
                *)
                    jat__log_warn "unknown init system: $initsys"
                    ;;
            esac
            ;;
        scan)
            mapfile -t services <<<"$(
                case $initsys in
                    systemd)
                        systemctl list-unit-files 2>/dev/null \
                          | grep '\<enabled\>' \
                          | grep -o '[^[:space:]]*[.]service\>' \
                          | rev \
                          | cut -d. -f2- \
                          | rev
                        ;;
                    sysvinit)
                        find /etc/init.d/ -perm /u=x,g=x,o=x -type f \
                          | sed s:/etc/init.d/::
                        ;;
                    *)
                        jat__log_warn "unknown init system: $initsys"
                        ;;
                esac \
                    | sort \
                    | uniq
            )"
            for sname in "${services[@]}"; do
                service "$sname" status &>/dev/null; sstate=$?
                echo "$sname:$sstate"
            done
            ;;
        describe)
            echo "state of services"
            ;;
        version)
            echo "0.0"
            ;;
    esac
}

__state__pushd() {
    #
    # Safe(r) pushd
    #
    local path=$1
    pushd "$path" >/dev/null && return 0
    jat__log_error "cannot chdir to: $path"
    return 3
}

__state__popd() {
    #
    # Safe(r) popd
    #
    local old
    local cur
    cur=$(dirs +0)
    old=$(dirs +1)
    popd >/dev/null && return 0
    jat__log_error "cannot popd: from: $cur to $old"
    return 3
}

__state__save1() {
    #
    # Save data from $Module into current directory
    #
    local modfn     # module's access function
    local moddesc   # ^^ description
    local modver    # ^^ version
    modfn="__state__m__$Module"
    type -t "$modfn"  >/dev/null || {
        jat__log_error "no such module: $Module"
        return 3
    }
    jat__pstarts -h "scan: $Module"
        moddesc=$($modfn describe)
        modver=$($modfn version)
        modver=${modver:-NONE}
        jat__log_info "module description: $moddesc" \
                      "module version: $modver"
        echo "$modver" > "$Module.meta.version"
        mkdir -p "$Module/diag"
        __state__pushd "$Module/diag" || return 3
            jat__cmd \
                "$modfn" diag
        __state__popd "$Module/diag" || return 3
        jat__cmd -o "$Module.out" -e "$Module.err" \
            "$modfn" scan
        jat_dump__file -S "$Module/diag/"*
        jat_dump__file "$Module.out"
        jat_dump__file -s -E "$Module.err"
    jat__pend
}

__state__starball() {
    #
    # Submit tarball of item id $2 of type $1
    #
    local type=$1
    local id=$2
    local tmp
    tmp=$(mktemp -t "__state__starball.$type.tmp.XXXXXXX")
    jat__cmd -h "create tarball: $STATE__DIR/$type/$id" \
        tar -czf "$tmp" -C "$STATE__DIR/$type" "$id"
    jat__submit "$tmp" "$id.tar.gz"
    rm "$tmp"
}

__state__usage() {
    #
    # Print usage message for $1
    #
    local short=$1
    local fn=state__$short
    local patts=()
    case $short in
        diff)       patts=("  $fn [-p PHASE] [-LA LABEL_A] [-LB LABEL_B] [-H HANDLER] SCAN_A SCAN_B") ;;
        diff_last2) patts=("  $fn [-p PHASE] [-LA LABEL_A] [-LB LABEL_B] [-H HANDLER]") ;;
    esac
    jat__log_error "usage:" "${patts[@]}"
}

#shellfu module-version=0.0.14
